#ifndef _UADE_MAIN_H_
#define _UADE_MAIN_H_

#include <uademsg.h>

struct uade_song {
  char playername[PATH_MAX];       // filename of eagleplayer
  char modulename[PATH_MAX];       // filename of song
  char scorename[PATH_MAX];        // filename of score file

  int set_subsong;
  int subsong;
  int force_by_default;
  int use_ntsc;
  int song_end_possible;
  int use_filter;

  int min_subsong;
  int max_subsong;
  int cur_subsong;
};

struct uade_command {
  int type;
  void *par;
  void *ret;
};

struct uade_local {
  void (*setup)(void (*interrupt)(void));
  void (*enqueue_song)(char *file);
  void (*get_next)(char *dst, int maxsize);
  int (*interaction)(struct uade_command *cmd, int wait_for);
  int (*song_list_empty)(void);
};

struct uade_remote {
  int (*setup)(char **slaveargs);
  void (*get_next)(struct uade_song *uade_song);
  void (*song_init_failure)(void);
  void (*post_init)(void);
  void (*report)(int type, void *data);
  void (*flush)(void);
  void (*write)(void *sndbuffer, int bytes);
};

void uade_audxdat(int nr, unsigned int v);
void uade_audxlch(int nr, unsigned int v);
void uade_audxlcl(int nr, unsigned int v);
void uade_audxlen(int nr, unsigned int v);
void uade_audxper(int nr, unsigned int v);
void uade_audxvol(int nr, unsigned int v);
void uade_aud_dma(unsigned int v);
void uade_aud_strike(int nr, unsigned int v);

void uade_change_subsong(int subs);
void uade_get_amiga_message(void);
void uade_option(int, char**); /* handles command line parameters */
void uade_prerun(void);
void uade_send_amiga_message(int msgtype);
void uade_set_automatic_song_end(int song_end_possible);
void uade_set_ntsc(int usentsc);
void uade_song_end(char *reason, int kill_it);
void uade_swap_buffer_bytes(void *data, int bytes);
void uade_test_sound_block(void *buf, int size); /* for silence detection */
void uade_vsync_handler(void);

extern int uade_remote_slave;
extern struct uade_remote *uade_remote;

extern int uade_soundfd;
extern int uade_usingoutpipe;
extern int uade_local_sound;
extern int uade_swap_output_bytes;

extern int uade_reboot;

extern int uade_do_panning;
extern float uade_pan_value;

extern int uade_time_critical;

extern int uade_debug;
extern int uade_filter_debug; /* just for debugging */
extern int uade_filter_counter; /* just for debugging */
extern int uade_filter_overruns; /* just for debugging */

#endif
